from typing import List, Tuple

from pyspark.sql import Column, DataFrame
import pyspark.sql.functions as F

from aporia.core.types.field import FieldType
from aporia.pyspark.training.pyspark_field_metadata import PySparkFieldMetadata


def _get_keys_of_dict_col(data: DataFrame, col: str) -> List[str]:
    # Inspired by https://mungingdata.com/pyspark/dict-map-to-multiple-columns/
    keys_df = data.select(F.explode(F.map_keys(F.col(col)))).distinct()
    return [row[0] for row in keys_df.collect()]


def expand_dicts_keys_attributes_to_regular_fields(
    data: DataFrame, dict_fields: List[PySparkFieldMetadata]
) -> List[Tuple[PySparkFieldMetadata, Column]]:
    """Expands dict fields to columns in a dataframe.

    Each attribute in the dict key will be expanded to an independent column.

    Args:
        data: the original dataframe
        dict_fields: A list of dict fields to expand

    Returns:
        List of (field metadata, corresponding dataframe column) tuples.
    """
    result: List[Tuple[PySparkFieldMetadata, Column]] = []

    for dict_field in dict_fields:
        keys = _get_keys_of_dict_col(data=data, col=dict_field.column)
        for key in keys:

            col_name = dict_field.column + "." + key

            key_field_metadata = PySparkFieldMetadata(
                column=col_name,
                name=dict_field.name,
                group=dict_field.group,
                type=FieldType.NUMERIC,
                key_at_dict=key,
            )

            col_query = F.col(dict_field.column).getItem(key).alias(col_name)

            result.append((key_field_metadata, col_query))

    return result



# Note: Don't import anything from aporia.training or
# aporia.pyspark here, it won't work without extra dependencies
import aporia.experimental  # noqa: F401
from .core.core_api import create_model, create_model_version, delete_model, init, shutdown
from .core.model_tags import add_model_tags, delete_model_tag, get_model_tags
from .core.types.data_source import (
    AthenaDataSource,
    BigQueryDataSource,
    GlueDataSource,
    JDBCDataSource,
    PostgresJDBCDataSource,
    S3DataSource,
    SnowflakeDataSource,
)
from .core.types.model import ModelColor, ModelIcon
from .core.utils import generate_features_schema_from_shape
from .model import Model

__all__ = [
    # Core
    "create_model",
    "create_model_version",
    "delete_model",
    "init",
    "shutdown",
    "add_model_tags",
    "delete_model_tag",
    "get_model_tags",
    "ModelColor",
    "ModelIcon",
    "generate_features_schema_from_shape",
    "AthenaDataSource",
    "BigQueryDataSource",
    "JDBCDataSource",
    "PostgresJDBCDataSource",
    "S3DataSource",
    "S3FileFormat",
    "SnowflakeDataSource",
    "GlueDataSource",
    # Inference
    "Model",
]



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Union


class S3FileFormat(Enum):
    """Supported file formats for S3 files."""

    PARQUET = "parquet"
    CSV = "csv"
    JSON = "json"
    DELTA = "delta"


class DataSourceType(Enum):
    """Supported data source types."""

    BIGQUERY = "bigquery"
    JDBC = "jdbc"
    POSTGRES_JDBC = "postgres-jdbc"
    ATHENA = "athena"
    SNOWFLAKE = "snowflake"
    GLUE = "glue"
    S3 = "s3"


class DataSource(ABC):
    """Base class for data sources."""

    def serialize(self) -> Dict[str, Any]:
        """Serialize data source to dict."""
        return {"type": self.type.value, "config": self.serialize_config()}

    @property
    @abstractmethod
    def type(self) -> DataSourceType:
        """Returns data source type."""
        ...

    @abstractmethod
    def serialize_config(self) -> Dict[str, Any]:
        """Serialize data source configuration to dict."""
        ...


class BaseSparkDataSource(DataSource):
    """Base class for spark-based data sources."""

    def __init__(
        self,
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
    ):
        """Initializes a BaseSparkDataSource.

        Args:
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
        """
        super().__init__()
        self.sample_size = sample_size
        self.select_expr = select_expr

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            "sample_size": self.sample_size,
            "select_expr": self.select_expr,
        }


class SparkDataSource(BaseSparkDataSource):
    """Base class for spark-based data sources using generic load API."""

    def __init__(
        self,
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
        read_options: Optional[Dict[str, Any]] = None,
    ):
        """Initializes a SparkDataSource.

        Args:
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
            read_options: Additional spark read options
        """
        super().__init__(sample_size=sample_size, select_expr=select_expr)
        self.read_options = read_options

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "read_options": self.read_options,
        }


class HiveDataSource(BaseSparkDataSource):
    """Base class for spark-based data sources using Hive API."""

    def __init__(
        self,
        query: str,
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
    ):
        """Initializes a HiveDataSource.

        Args:
            query: SQL query to read data from the database
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
        """
        super().__init__(sample_size=sample_size, select_expr=select_expr)
        self.query = query

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "query": self.query,
        }


class JDBCDataSource(SparkDataSource):
    """Generic JDBC data source."""

    def __init__(
        self,
        url: str,
        query: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        sample_size: float = 1,
        select_expr: Optional[List[str]] = None,
        read_options: Optional[Dict[str, Any]] = None,
    ):
        """Initializes a JDBCDataSource.

        Args:
            url: Database connection URL
            query: SQL query to read data from the database
            user: Database user
            password: Database password
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
            read_options: Additional spark read options
        """
        super().__init__(
            sample_size=sample_size, select_expr=select_expr, read_options=read_options
        )
        self.url = url
        self.query = query
        self.user = user
        self.password = password

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.JDBC

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "url": self.url,
            "query": self.query,
            "user": self.user,
            "password": self.password,
        }


class PostgresJDBCDataSource(JDBCDataSource):
    """Postgres (via JDBC) data source."""

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.POSTGRES_JDBC


class AthenaDataSource(JDBCDataSource):
    """AWS Athena data source."""

    def __init__(
        self,
        url: str,
        query: str,
        s3_output_location: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        sample_size: float = 1,
        select_expr: Optional[List[str]] = None,
        read_options: Optional[Dict[str, Any]] = None,
    ):
        """Initializes a JDBCDataSource.

        Args:
            url: Database connection URL
            query: SQL query to read data from the database
            s3_output_location: Path to S3 bucket for storing query results
            user: Database user
            password: Database password
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
            read_options: Additional spark read options
        """
        super().__init__(
            url=url,
            query=query,
            user=user,
            password=password,
            sample_size=sample_size,
            select_expr=select_expr,
            read_options=read_options,
        )
        self.s3_output_location = s3_output_location

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.ATHENA

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "s3_output_location": self.s3_output_location,
        }


class BigQueryDataSource(SparkDataSource):
    """BigQuery data source."""

    def __init__(
        self,
        credentials_base64: str,
        table: str,
        dataset: Optional[str] = None,
        project: Optional[str] = None,
        parent_project: Optional[str] = None,
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
        read_options: Optional[Dict[str, Any]] = None,
    ):
        """Initializes a SparkDataSource.

        Args:
            credentials_base64: Base64 encoded JSON string containing GCP service account details.
            table: Table to query
            dataset: Dataset to query
            project: Project name
            parent_project: Parent project name
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
            read_options: Additional spark read options
        """
        super().__init__(
            sample_size=sample_size, select_expr=select_expr, read_options=read_options
        )
        self.credentials_base64 = credentials_base64
        self.table = table
        self.dataset = dataset
        self.project = project
        self.parent_project = parent_project

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.BIGQUERY

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "credentials_base64": self.credentials_base64,
            "table": self.table,
            "dataset": self.dataset,
            "project": self.project,
            "parent_project": self.parent_project,
        }


class SnowflakeDataSource(SparkDataSource):
    """Snowflake data source."""

    def __init__(
        self,
        url: str,
        query: str,
        user: str,
        password: str,
        database: str,
        schema: str,
        warehouse: Optional[str] = None,
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
        read_options: Optional[Dict[str, Any]] = None,
    ):
        """Initializes a SnowflakeDataSource.

        Args:
            url: The full Snowflake URL of your instance
            query: SQL query to read data from the database
            user: Username for database connection
            password: Password for database connection
            database: Database name
            schema: Schema name
            warehouse: The default virtual warehouse to use
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
            read_options: Additional spark read options
        """
        super().__init__(
            sample_size=sample_size, select_expr=select_expr, read_options=read_options
        )
        self.url = url
        self.query = query
        self.user = user
        self.password = password
        self.database = database
        self.schema = schema
        self.warehouse = warehouse

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.SNOWFLAKE

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "url": self.url,
            "query": self.query,
            "user": self.user,
            "password": self.password,
            "database": self.database,
            "sfschema": self.schema,
            "warehouse": self.warehouse,
        }


class GlueDataSource(HiveDataSource):
    """Glue data source."""

    def __init__(
        self,
        query: str,
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
    ):
        """Initializes a GlueDataSource.

        Args:
            query: SQL query to read data from the database
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
        """
        super().__init__(sample_size=sample_size, select_expr=select_expr, query=query)

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.GLUE


class S3DataSource(SparkDataSource):
    """S3 data source."""

    def __init__(
        self,
        object_path: str,
        object_format: Union[S3FileFormat, str],
        sample_size: float = 1.0,
        select_expr: Optional[List[str]] = None,
        read_options: Optional[Dict[str, Any]] = None,
    ):
        """Initializes a S3DataSource.

        Args:
            object_path: The path in S3 to the object, excluding s3 prefix (e.g: bucket-name/file.parquet)
            object_format: Type of the input file (Parquet, CSV, JSON, etc)
            sample_size: Fraction of data to sample
            select_expr: Select expressions to apply to the dataframe after reading
            read_options: Additional spark read options
        """
        super().__init__(
            sample_size=sample_size, select_expr=select_expr, read_options=read_options
        )
        self.object_path = object_path
        self.object_format = S3FileFormat(object_format)

    @property
    def type(self) -> DataSourceType:
        """See base class."""
        return DataSourceType.S3

    def serialize_config(self) -> Dict[str, Any]:
        """See base class."""
        return {
            **super().serialize_config(),
            "object_path": self.object_path,
            "object_format": self.object_format.value,
        }



from .pyspark_calculate_fields import calculate_training_fields
from .pyspark_create_fields import create_fields_metadata
from .pyspark_field_metadata import PySparkFieldMetadata

__all__ = ["PySparkFieldMetadata", "create_fields_metadata", "calculate_training_fields"]



from typing import cast, Dict, List, Optional

import numpy as np
from pandas import DataFrame, Series
from sklearn.metrics import confusion_matrix, mean_absolute_error, mean_squared_error

from aporia.core.consts import TRAINING_BIN_COUNT
from aporia.core.errors import AporiaError
from aporia.core.types.field import FieldType
from .api.log_training import FieldTrainingData


def calculate_dataframe_training_data(
    data: DataFrame, fields_schema: Dict[str, FieldType], labels_data: Optional[DataFrame] = None
) -> List[FieldTrainingData]:
    """Calculates training data for all fields in a DataFrame.

    Args:
        data: Raw training data
        fields_schema: Fields schema for the category that is being calculated.
        labels_data: Raw labels of training data. This field is relevant only when the data
            is the predictions so we could calculate performance metrics. For calculating
            statistic aggregation on labels, pass the labels in the data arg, and leave
            this argument empty.

    Returns:
        Training data for all fields in the dataframe
    """
    training_data = []
    for field_name, field_data in data.items():
        # Ignore fields that are not defined in the model version schema
        if field_name not in fields_schema:
            continue

        field_label_data = labels_data.loc[:, field_name] if labels_data is not None else None
        training_data.extend(
            calculate_training_data(
                field_name=field_name,
                field_data=field_data,
                field_type=fields_schema[field_name],
                field_label_data=field_label_data,
            )
        )

    return training_data


def calculate_training_data(
    field_name: str,
    field_data: Series,
    field_type: FieldType,
    field_label_data: Optional[Series] = None,
) -> List[FieldTrainingData]:
    """Calculates training data for a single field.

    Args:
        field_name: Field name
        field_data: Field data
        field_type: Field type
        field_label_data: The label data of field. Relevant only when the field if
            a prediction field. Used to calculate the performance metrics.

    Returns:
        Field training data.
    """
    # We currently don't support datetime, vector, text & dict training data
    if field_type in [FieldType.DATETIME, FieldType.VECTOR, FieldType.TEXT, FieldType.TENSOR]:
        return [
            _calculate_training_data_without_histogram(field_name=field_name, field_data=field_data)
        ]
    elif field_type == FieldType.NUMERIC:
        return [
            _calculate_numeric_training_data(
                field_name=field_name, field_data=field_data, field_label_data=field_label_data
            )
        ]
    elif field_type == FieldType.DICT:
        return _calculate_dict_training_data(field_name=field_name, field_data=field_data)
    elif field_type in [FieldType.BOOLEAN, FieldType.STRING, FieldType.CATEGORICAL]:
        return [
            _calculate_categorical_training_data(
                field_name=field_name,
                field_data=field_data,
                field_type=field_type,
                field_label_data=field_label_data,
            )
        ]

    raise AporiaError("Unsupported field type {} of field {}".format(field_type.value, field_name))


def _calculate_dict_training_data(field_name: str, field_data: Series) -> List[FieldTrainingData]:
    # Each row in field_data is expected to be a dict: {"a": 0.5, "b": 0.2}
    # By using .apply(Series), we can expand each key to a column in a dataframe
    expanded_dict_data = field_data.apply(Series)

    training_data = []
    for key, key_data in expanded_dict_data.items():
        training_data.append(
            _calculate_numeric_training_data(field_name=field_name, key=key, field_data=key_data)
        )

    return training_data


def _calculate_training_data_without_histogram(
    field_name: str, field_data: Series
) -> FieldTrainingData:
    valid_values = field_data[field_data.notnull()]

    return FieldTrainingData(
        field_name=field_name,
        num_samples=len(valid_values),
        num_missing_values=len(field_data) - len(valid_values),
    )


def _calculate_categorical_training_data(
    field_name: str,
    field_data: Series,
    field_type: FieldType,
    field_label_data: Optional[Series] = None,
) -> FieldTrainingData:
    valid_values = field_data[field_data.notnull()]

    # for performance we want all the values that both predictions and labels are not null
    valid_performance_values = None
    valid_label_values = None
    if field_label_data is not None:
        no_none_preformence_mask = field_data.notnull() & field_label_data.notnull()
        valid_performance_values = field_data[no_none_preformence_mask]
        valid_label_values = field_label_data[no_none_preformence_mask]

    tn, fp, fn, tp, recall_score, precision_score, accuracy, f1_score = 8 * [None]
    if field_type == field_type.BOOLEAN:
        valid_values = valid_values.astype("bool")
        if field_label_data is not None:
            valid_performance_values = cast(Series, valid_performance_values)
            valid_label_values = cast(Series, valid_label_values)
            valid_performance_values = valid_performance_values.astype("bool")
            valid_label_values = valid_label_values.astype("bool")
            tn, fp, fn, tp = confusion_matrix(valid_label_values, valid_performance_values).ravel()
            recall_score = tp / (tp + fn) if (tp + fn) > 0 else None
            precision_score = tp / (tp + fp) if (tp + fp) > 0 else None
            accuracy = (tp + tn) / (tp + tn + fp + fn) if (tp + tn + fp + fn) > 0 else None
            f1_score = 2 * tp / (2 * tp + fp + fn) if (2 * tp + fp + fn) > 0 else None

    bins, counts = np.unique(valid_values, return_counts=True)

    top_frequency = counts.max()
    top_index = np.where(counts == top_frequency)[0][0]
    top = bins[top_index]

    # Note: There is a possible edge case here in which a user passes an infinite value
    # as one of the categories. We chose not to count those values at the moment, since
    # most numpy functions don't handle str and bool well, which would force us to split
    # this function up for each field type.
    return FieldTrainingData(
        field_name=field_name,
        bins=bins.tolist(),
        counts=counts.tolist(),
        num_samples=len(valid_values),
        num_missing_values=len(field_data) - len(valid_values),
        num_unique_values=len(bins),
        tn=tn,
        fp=fp,
        fn=fn,
        tp=tp,
        recall_score=recall_score,
        precision_score=precision_score,
        accuracy=accuracy,
        f1_score=f1_score,
        top=top,
        top_frequency=top_frequency,
    )


def _calculate_numeric_training_data(
    field_name: str,
    field_data: Series,
    key: Optional[str] = None,
    field_label_data: Optional[Series] = None,
) -> FieldTrainingData:
    # Cast everything to float and normalize null values (some operations don't work on pd.NA)
    field_data = field_data.fillna(np.nan).astype(float)  # type: ignore
    if field_label_data is not None:
        field_label_data = field_label_data.fillna(np.nan).astype(float)  # type: ignore

    # Filter out infinite and NaN values (isfinite is False for NaN)
    finite_values = field_data[np.isfinite(field_data)]
    counts, bins = np.histogram(finite_values, bins=TRAINING_BIN_COUNT)

    # np.median and np.average print redundant runtime warnings when the data is empty
    median = np.median(finite_values) if finite_values.size > 0 else np.nan
    average = np.average(finite_values) if finite_values.size > 0 else np.nan

    mse = None
    rmse = None
    mae = None
    if field_label_data is not None:
        # for performance we want all the values that both predictions and labels are not null
        finite_performance_values = field_data[
            np.isfinite(field_label_data) & np.isfinite(field_data)
        ]
        finite_performance_label_values = field_label_data[
            np.isfinite(field_label_data) & np.isfinite(field_data)
        ]

        mse = mean_squared_error(
            finite_performance_label_values, finite_performance_values, squared=True
        )
        rmse = mean_squared_error(
            finite_performance_label_values, finite_performance_values, squared=False
        )
        mae = mean_absolute_error(finite_performance_label_values, finite_performance_values)

    return FieldTrainingData(
        field_name=field_name,
        key=key,
        bins=bins.tolist(),
        counts=counts.tolist(),
        min=np.min(finite_values),
        max=np.max(finite_values),
        sum=np.sum(finite_values),  # type: ignore
        median=median,
        average=average,
        std=np.std(finite_values),  # type: ignore
        variance=np.var(finite_values),  # type: ignore
        num_samples=len(finite_values),
        num_missing_values=np.count_nonzero(np.isnan(field_data)),
        num_posinf_values=np.count_nonzero(np.isposinf(field_data)),
        num_neginf_values=np.count_nonzero(np.isneginf(field_data)),
        num_zero_values=np.count_nonzero(finite_values == 0),
        mse=mse,
        rmse=rmse,
        mae=mae,
    )



from typing import Any, Dict, Optional

from aporia.core.http_client import HttpClient
from aporia.core.types.data_source import DataSource

DEFAULT_CONNECTION_TIMEOUT_SECONDS = 60 * 10


def _get_query_data(
    dataset: str,
    environment: str,
    data_source: DataSource,
    id_column: str,
    timestamp_column: str,
    predictions: Optional[Dict[str, str]] = None,
    features: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    raw_inputs: Optional[Dict[str, str]] = None,
) -> Dict[str, Any]:
    data = {
        "name": dataset,
        "data_source": data_source.serialize(),
        "sync": {
            "enabled": True,
            "environment": environment,
        },
        "id_column": id_column,
        "timestamp_column": timestamp_column,
    }

    if predictions is not None:
        data["predictions"] = predictions

    if features is not None:
        data["features"] = features

    if labels is not None:
        data["labels"] = labels

    if raw_inputs is not None:
        data["raw_inputs"] = raw_inputs

    return data


async def connect_dataset(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    dataset: str,
    environment: str,
    data_source: DataSource,
    id_column: str,
    timestamp_column: str,
    predictions: Optional[Dict[str, str]] = None,
    features: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    raw_inputs: Optional[Dict[str, str]] = None,
    http_timeout_seconds: Optional[int] = None,
):
    """Connect to a dataset.

    Args:
        http_client: HTTP client.
        model_id: Model id.
        model_version: Model version.
        dataset: Dataset name.
        environment: Environment.
        data_source: Data source object (for example, `SparkSQLDataSource`).
        id_column: Name of the ID column.
        timestamp_column: Name of the timestamp column.
        predictions: Predictions -> columns mappping. Defaults to None.
        features: Features -> columns mappping. Defaults to None.
        labels: Labels -> columns mappping. Defaults to None.
        raw_inputs: Raw inputs -> columns mappping. Defaults to None.
        http_timeout_seconds: HTTP timeout in seconds. Defaults to 10 minutes.
    """
    http_timeout_seconds = (
        http_timeout_seconds
        if http_timeout_seconds is not None
        else DEFAULT_CONNECTION_TIMEOUT_SECONDS
    )

    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/datasets",
        data=_get_query_data(
            dataset=dataset,
            environment=environment,
            data_source=data_source,
            id_column=id_column,
            timestamp_column=timestamp_column,
            features=features,
            predictions=predictions,
            labels=labels,
            raw_inputs=raw_inputs,
        ),
        timeout=http_timeout_seconds,
    )


async def set_datasets_sync_enabled(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    enabled: bool,
):
    """Set the synchronization status of the model version.

    Args:
        http_client: HTTP client.
        model_id: Model id.
        model_version: Model version.
        enabled: Dataset name.
    """
    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/datasets_sync?enabled={enabled}", data={}
    )



from collections import defaultdict
from json import dumps, loads
import logging
from typing import Any, Dict, List, Optional, Tuple

from aporia.core.logging_utils import LOGGER_NAME
from aporia.pyspark.training.pyspark_field_metadata import PySparkFieldMetadata
from aporia.training.api.log_training import FieldTrainingData

logger = logging.getLogger(LOGGER_NAME)


def get_field_identifier(field_group: str, field_name: str, dict_key: Optional[str] = None) -> str:
    """Returns a string that can be used to identify a field from the model."""
    hash_data: Dict = {"group": field_group, "name": field_name}
    if dict_key is not None:
        hash_data["key_at_dict"] = dict_key
    return dumps(hash_data)


def _parse_field_identifier(field_identifier_str: str) -> Tuple[str, str, Optional[str]]:
    hash_data: Dict = loads(field_identifier_str)
    group = hash_data["group"]
    name = hash_data["name"]
    dict_key = hash_data.get("key_at_dict", None)
    return group, name, dict_key


def create_alias(aggregation_name: str, field: PySparkFieldMetadata) -> str:
    """Create the alias string for the pyspark quires."""
    alias_data = {
        "aggregation_name": aggregation_name,
        "group": field.group,
        "column": field.column,
        "name": field.name,
    }
    if field.key_at_dict is not None:
        alias_data["key_at_dict"] = field.key_at_dict
    return dumps(alias_data)


def _parse_alias(alias_str: str) -> Tuple[str, str]:
    """Parse an alias str."""
    logger.debug("parsing aggregation alias: " + alias_str)
    alias_data: Dict[str, str] = loads(alias_str)
    field_identifier = get_field_identifier(
        field_group=alias_data["group"],
        field_name=alias_data["name"],
        dict_key=alias_data.get("key_at_dict", None),
    )
    return field_identifier, alias_data["aggregation_name"]


def parse_aggregation_results_to_field_training_data_by_group(
    aggregate_results_dict: dict,
) -> Dict[str, List[FieldTrainingData]]:
    """Parses aggregation results.

    Args:
        aggregate_results_dict: Aggregation results

    Returns:
        Field training data for each field group
    """
    field_identifier_to_metrics: Dict[str, Dict[str, Any]] = defaultdict(lambda: {})
    for agg_alias, agg_result in aggregate_results_dict.items():
        field_identifier, aggregation_name = _parse_alias(agg_alias)
        field_identifier_to_metrics[field_identifier][aggregation_name] = agg_result

    fields_training_data_by_group: Dict[str, List[FieldTrainingData]] = {}
    for field_identifier, metrics in field_identifier_to_metrics.items():
        group, name, dict_key = _parse_field_identifier(field_identifier)
        if group not in fields_training_data_by_group:
            fields_training_data_by_group[group] = []
        if dict_key is None:
            fields_training_data_by_group[group].append(
                FieldTrainingData(field_name=name, **metrics)
            )
        else:
            fields_training_data_by_group[group].append(
                FieldTrainingData(field_name=name, key=dict_key, **metrics)
            )
    return fields_training_data_by_group



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aporia',
 'aporia.core',
 'aporia.core.api',
 'aporia.core.types',
 'aporia.experimental',
 'aporia.inference',
 'aporia.inference.api',
 'aporia.inference.types',
 'aporia.pandas',
 'aporia.pyspark',
 'aporia.pyspark.experimental',
 'aporia.pyspark.training',
 'aporia.training',
 'aporia.training.api']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.0,<4.0.0',
 'certifi>=2022.6.15,<2023.0.0',
 'orjson>=3.6.4,<4.0.0',
 'tenacity>=7.0,<9.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.5.0,<2.0.0'],
 ':python_version >= "3.7" and python_version < "3.8"': ['numpy>=1.15,<1.22',
                                                         'scikit-learn>=1.0,<1.1'],
 ':python_version >= "3.8" and python_version < "4.0"': ['numpy>=1.17.4,<2.0.0',
                                                         'scikit-learn>=1.0,<2.0'],
 'all': ['pyspark>=3.0,<4.0'],
 'all:python_version >= "3.7" and python_version < "3.8"': ['pandas>=1.0,<1.4',
                                                            'pandas>=1.0,<1.4'],
 'all:python_version >= "3.8" and python_version < "4.0"': ['pandas>=1.0,<2.0',
                                                            'pandas>=1.0,<2.0'],
 'pandas:python_version >= "3.7" and python_version < "3.8"': ['pandas>=1.0,<1.4',
                                                               'pandas>=1.0,<1.4'],
 'pandas:python_version >= "3.8" and python_version < "4.0"': ['pandas>=1.0,<2.0',
                                                               'pandas>=1.0,<2.0'],
 'pyspark': ['pyspark>=3.0,<4.0'],
 'training:python_version >= "3.7" and python_version < "3.8"': ['pandas>=1.0,<1.4',
                                                                 'pandas>=1.0,<1.4'],
 'training:python_version >= "3.8" and python_version < "4.0"': ['pandas>=1.0,<2.0',
                                                                 'pandas>=1.0,<2.0']}

setup_kwargs = {
    'name': 'aporia',
    'version': '2.18.0',
    'description': 'Aporia SDK',
    'long_description': '# Aporia SDK\n\n## Testing\n\nTo run the tests, first install the library locally:\n```\npip install ".[all]" --upgrade\n```\n\nThen run the tests using `pytest`:\n```\npytest -v\n```\n\nIf you don\'t have Spark installed, skip the pyspark tests:\n```\npytest -v --ignore=tests/pyspark\n```\n',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/aporia-ai/sdk',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections import abc, OrderedDict
from typing import Any, Dict, Iterable, Optional, Union

import numpy as np
import pandas as pd

from aporia.core.context import get_context
from aporia.core.errors import AporiaError, handle_error
from aporia.core.types.field import FieldSchema, FieldType, FieldValue


def _dtype_to_field_type(dtype: str) -> Optional[FieldSchema]:
    # This is based on https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html#numpy.dtype.kind
    DTYPE_TO_FIELD_TYPE = {
        "b": FieldType.BOOLEAN,
        "i": FieldType.NUMERIC,
        "u": FieldType.NUMERIC,
        "f": FieldType.NUMERIC,
        "M": FieldType.DATETIME,
        "O": FieldType.STRING,
        "S": FieldType.STRING,
        "U": FieldType.STRING,
    }
    field_type = DTYPE_TO_FIELD_TYPE.get(dtype)
    if field_type is None:
        return None
    return {"type": field_type.value}


NUMERIC_DTYPES = ["i", "u", "f"]
STRING_DTYPES = ["O", "S", "U"]
STRING_UNIQUE_RATIO = 0.25
MIN_UNIQUE_VALUES_FOR_TEXT = 50


def _infer_string_data_for_text_or_string(data: pd.Series) -> FieldSchema:
    """Convert a numpy/pandas of series contains only strings to a FieldType.

    Args:
        data: pandas data series

    Returns:
        FieldType that matches the dtype
    """
    # We need to decided if a string type is str or text
    # We check if there are more than 50 unique values
    # which are more than 25% of the total values
    _, unique_counts = np.unique(data.values, return_counts=True)
    uniques = len(unique_counts)
    if uniques > MIN_UNIQUE_VALUES_FOR_TEXT and uniques > data.shape[0] * STRING_UNIQUE_RATIO:
        return {"type": FieldType.TEXT.value}

    return {"type": FieldType.STRING.value}


def _replace_na_type(data: pd.Series) -> pd.Series:
    """Replace NAType with NA (since they are not removed by dropna).

    Args:
        data: pandas data series

    Returns:
        The data series where the NAType objects are replaced with NA
    """
    result = data
    if data.apply(lambda x: x == pd._libs.missing.NAType).any():
        result = data.copy()
        for i, value in result.iteritems():
            if value == pd._libs.missing.NAType:
                result.at[i] = pd.NA
    return result


def _infer_object_dtype_column(data: pd.Series) -> Optional[FieldSchema]:
    """Attempts to convert a numpy/pandas series of object dtype to a FieldType.

    Args:
        data: pandas data series

    Returns:
        FieldType that matches the dtype, or None if conversion is impossible
    """
    if data.size == 0:
        return None
    elif data.apply(lambda x: isinstance(x, dict)).all():
        return {"type": FieldType.DICT.value}
    elif data.apply(lambda x: isinstance(x, str)).all():
        return _infer_string_data_for_text_or_string(data=data)
    elif data.apply(lambda x: isinstance(x, abc.Iterable)).all():
        return {"type": FieldType.TENSOR.value, "dimensions": [len(data.values[0])]}
    return None


def infer_field_schema_from_dtype_and_data(
    dtype: np.dtype, data: Optional[pd.Series]
) -> Optional[FieldSchema]:
    """Attempts to convert a numpy/pandas dtype to a FieldType.

    Args:
        dtype: Dtype to convert
        data: pandas data series

    Returns:
        FieldType that matches the dtype, or None if conversion is impossible
    """
    if isinstance(dtype, pd.api.types.CategoricalDtype):
        category_type = _dtype_to_field_type(dtype.categories.dtype.kind)
        # We only support categorical fields with numeric categories
        if category_type is None:
            return None
        if category_type["type"] == FieldType.NUMERIC.value:
            return {"type": FieldType.CATEGORICAL.value}

        return category_type

    if dtype.kind == "O":
        return _infer_object_dtype_column(data=data)

    if data is not None and not data.empty:
        if dtype.kind in NUMERIC_DTYPES and isinstance(data.values[0], abc.Iterable):
            arr: Any = data.values[0]
            return {"type": FieldType.TENSOR.value, "dimensions": [len(arr)]}

        # We need to decided if a string type is str or text
        # We check if there are more than 50 unique values
        # which are more than 25% of the total values
        if dtype.kind in STRING_DTYPES and len(data.shape) > 0:

            return _infer_string_data_for_text_or_string(data=data)

    return _dtype_to_field_type(dtype.kind)  # type: ignore


def pandas_to_dict(data: Union[pd.DataFrame, pd.Series]) -> Optional[Dict[str, FieldValue]]:
    """Converts a pandas DataFrame or Series to a dict for log_* functions.

    Args:
        data: DataFrame or Series to convert.

    Returns:
        The data converted to a dict, mapping field names to their values

    Notes:
        * data must contain column names that match the fields defined in create_model_version
        * If data is a DataFrame, it must contain exactly one row
    """
    context = None
    try:
        context = get_context()

        if isinstance(data, pd.Series):
            return data.fillna(np.nan).to_dict()  # type: ignore
        elif isinstance(data, pd.DataFrame):
            num_rows, _ = data.shape
            if num_rows > 1:
                raise AporiaError("cannot convert DataFrame with more than 1 row")

            return data.iloc[0].fillna(np.nan).to_dict()
        else:
            raise AporiaError("data must be a pandas DataFrame or Series")

    except Exception as err:
        handle_error(
            message_format="Converting pandas data to dict failed, {}",
            verbose=False if context is None else context.config.verbose,
            throw_errors=False if context is None else context.config.throw_errors,
            debug=False if context is None else context.config.debug,
            original_exception=err,
        )

    return None


def infer_schema_from_dataframe(data: pd.DataFrame) -> Optional[Dict[str, Dict]]:
    """Infers model version schema from a pandas DataFrame or Series.

    Field names and types are inferred from column names and types.

    Args:
        data: pandas DataFrame or Series

    Returns:
        A schema describing the data, as required by the create_model_version function.

    Notes:
        * The field types are inferred using the following logic, based on the column dtypes:
            * dtype="category" with numeric (integer of float) categories -> categorical field
            * dtype="category" with non-numeric categories -> See rules below
            * Array of numeric values (integer or float) -> vector field
            * dtype="bool" -> boolean field
            * dtypes that represent signed/unsigned integers and floating point numbers -> numeric field
            * dtype is "string", "unicode", "object" -> string field
            * dtype is "string", "unicode", "object", more than 50 values which more than
                25% of them are unique -> text field
            * dtype is any datetime type (with or without timezone) -> datetime field
        * If data contains a column with a type that doesn't match any of the rules
          described above, an error will be raised.
    """
    if not isinstance(data, pd.DataFrame):
        raise AporiaError(f"cannot infer schema from {type(data)}, data must be a pandas DataFrame")

    schema = OrderedDict()
    not_inferable_columns = []
    for column_name, values in data.items():
        column_name = str(column_name)

        values_without_nulls = _replace_na_type(values)
        values_without_nulls = values_without_nulls.dropna().infer_objects()

        column_type = infer_field_schema_from_dtype_and_data(
            values_without_nulls.dtype, values_without_nulls
        )
        if column_type is None:
            not_inferable_columns.append(column_name)
        else:
            schema[column_name] = column_type

    if len(not_inferable_columns) > 0:
        raise AporiaError(
            f"Could not infer the type of columns {not_inferable_columns}. "
            "To fix, make sure to set column dtype. For example "
            f"df['{not_inferable_columns[0]}'] = df['{not_inferable_columns[0]}'].astype(np.int32)"
        )

    return schema


def create_null_dataframe(size: int, columns: Iterable[str]) -> pd.DataFrame:
    """Creates a dataframe with nulls for each column.

    Args:
        size: Number of rows to create.
        columns: Iterable of column names.

    Returns:
        Dataframe with nulls for each column.
    """
    return pd.DataFrame({column: [None for _ in range(size)] for column in columns})
